==> mcet-1.0.x-dev.tar.gz: mcet/src/Query/Query.php <==
<?php

/**
 * @file
 * Contains \Drupal\external_entities\Entity\Query\External\Query.
 */

namespace Drupal\mongo_content_entity_type\Query;

use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\Query\QueryBase;
use Drupal\Core\Entity\Query\QueryInterface;
use Drupal\mongo_content_entity_type\Entity\MongoContentEntityBaseInterface;
use Drupal\mongodb\DatabaseFactory;
use MongoDB\Model\BSONDocument;
use MongoDB\Model\IndexInfo;
use MongoDB\Operation\Explain;
use MongoDB\Operation\Explainable;
use Symfony\Component\Serializer\Serializer;
use MongoDB\Driver\Exception\CommandException;

/**
 * The SQL storage entity query class.
 */
class Query extends QueryBase implements QueryInterface {

  /** @var \Drupal\mongodb\DatabaseFactory $mongodbFactory */
  public $factory;

  /** @var MongoDb\Database Database */
  public $database;

  /** @var \MongoDB\Collection */
  private $collection = null;

  /** @var \Symfony\Component\Serializer\Serializer $serializer */
  public $serializer;

  public $filters = [];
  protected $options = [];
  protected $projection = [];

  public $fieldDefinitions = [];

  /** @var MongoContentEntityBaseInterface null */
  public $entityInstance = null;
  public $entity_id = null;
  public $bundle = null;

  public $returnDocuments = FALSE;

  /**
   * Constructs a query object.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type definition.
   * @param string $conjunction
   *   - AND: all of the conditions on the query need to match.
   * @param array $namespaces
   *   List of potential namespaces of the classes belonging to this query.
   * @param DatabaseFactory $factory
   *   List of potential namespaces of the classes belonging to this query.*
   */
  public function __construct(EntityTypeInterface $entity_type, $conjunction, array $namespaces, DatabaseFactory $factory, Serializer $serializer) {
    parent::__construct($entity_type, $conjunction, $namespaces);
    $this->factory = $factory;
    $this->serializer = $serializer;
  }


  /**
   * {@inheritdoc}
   */
  public function aggregate($stage, $logic, $langcode = NULL, &$alias = NULL) {
    $this->aggregate[($stage . "_" . sizeof($this->aggregate))] = [
      'stage' => $stage,
      'logic' => $logic,
      'langcode' => $langcode,
    ];

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function execute() {
    return $this
      ->prepare()
      ->compile()
      ->addSort()
      ->result();
  }

  /**
   * @param Explainable $operation
   * @return array|\MongoDB\Collection|object
   */
  public function explain($operation) {
    return $this->getCollection()->explain($operation, $this->options);
  }

  /**
   * @param bool $flag
   * @return $this
   */
  public function setExplain(bool $flag, $verbosity = Explain::VERBOSITY_QUERY): QueryAggregate {
    $this->options['explain'] = $flag;
    $this->options['verbosity'] = $verbosity;
    return $this;
  }

  public function setProjection(array $projection): Query {
    $this->projection = $projection;
    return $this;
  }

  /**
   * @param $filter
   * @param array|null $options
   * @return int
   */
  public function deleteMany($filter, array $options = []) {
    $result = $this->getCollection()->deleteMany($filter, $options);
    return $result->getDeletedCount();
  }

  /**
   * @param array|null $options
   * @return array|object
   */
  public function drop(array $options = []) {
    return $this->getCollection()->drop($options);
  }

  protected function addSort() {
    if ($this->count) {
      $this->sort = [];
    }

    $sort = [];
    if ($this->sort) {
      foreach ($this->sort as $key => $data) {
        $sort[$data['field']] = $data['direction'] === 'DESC' ? -1 : 1;
      }
    }

    $this->options = array_merge($this->options, [
      'sort' => $sort
    ]);

    return $this;
  }

  public function fetchDocuments() {
    $this->returnDocuments = TRUE;
//    $this->projection = $projection;
    return $this;
  }

  protected function prepare() {
    $this->setOptions();

    $this->entity_id = $this->getEntityTypeId();
    $this->bundle = $this->getEntityTypeId();

    $this->fieldDefinitions = \Drupal::entityManager()->getFieldDefinitions($this->entity_id, $this->bundle);

    $entityClass = $this->entityType->getClass();
    $this->entityInstance = new $entityClass([], $this->entity_id, $this->bundle);

    if (method_exists($this->entityInstance, 'getIndexes')) {
      $this->createIndexes();
    }

    return $this;
  }

  protected function createIndexes() {
    $cid = $this->getEntityTypeId() . '_mongo_indexed';
    $cache = \Drupal::cache()->get($cid);
    if ($cache)
      return;

    $requiredIndexes = $this->entityInstance->getIndexes();
    if (is_null($requiredIndexes))
      return;

    $cursor = $this->getCollection()->listIndexes();

    /** @var IndexInfo[] $indexes */
    $indexes = [];
    foreach ($cursor as $index) {
      $indexes[] = $index;
    }

    try {
      foreach ($requiredIndexes as $type => $data) {
        // Produces a single text index as per mongo specifications
        if ($type === 'string_indexes') {
          $this->_indexTextFields($data, $indexes);
        }

        // Produces multiple index per field
        if ($type === 'numeric_indexes') {
          $this->_indexNumericFields($data, $indexes);
        }
      }
    } catch (CommandException $e) {
      _wefiq_log_message("Error on indexes: " . $e->getMessage());
    }

    // TODO:: enable this
    \Drupal::cache()->set($cid, 1);
  }

  /**
   * @param $fields
   * @param \MongoDB\Model\IndexInfo[] $indexes
   */
  private function _indexNumericFields($fields, $indexes) {
    $newIndexes = [];
    foreach ($fields as $field_name => $field_values) {
      $existingIndexes = array_filter(array_map(function ($index) use ($field_name, $field_values) {
        /** @var \MongoDB\Model\IndexInfo $index */

        $fields = $index->getKey();
        $field_keys = array_keys($fields);
        if (in_array($field_name, $field_keys)) {
          $this->getCollection()->dropIndex($index);
          return $index;
        }
      }, $indexes), function ($value) {
        return !is_null($value);
      });

      $newIndexes[] = array_merge([
        'key' => $this->_buildNumericIndexFieldValue($field_name, $field_values)
      ]);
    }

    // drop all indexes which don't exist anymore
    foreach ($indexes as $index) {
      // TODO:: delete old indexes which don't exit anymore
    }

    if (!empty($newIndexes)) {
      $this->getCollection()->createIndexes($newIndexes);
    }
  }

  private function _buildNumericIndexFieldValue($name, $value) {
    $fieldValue = array_merge([
      $name => isset($value['value']) ? $value['value'] : 1,
    ], isset($value['extra']) ? $value['extra'] : []);

    return $fieldValue;
  }

  private function _indexTextFields($data, $indexes) {
    /** @var \MongoDB\Model\IndexInfo $textIndex */
    $textIndex = array_filter(array_map(function ($index) use ($data, $indexes) {
      /** @var \MongoDB\Model\IndexInfo $index */

      $keys = array_keys($index->getKey());

      // TODO: check it in more detail
      if (in_array($keys[0], ['_fts', '_ftx'])) {
        return $index;
      }
    }, $indexes), function ($value) {
      return !is_null($value);
    });

    if (!empty($textIndex)) {
      $textIndex = array_values($textIndex)[0];
    }

    $textFields = [];
    if (!empty($textIndex)) {
      $textFields = $textIndex->offsetGet('weights');
    }

    $newIndexes = [];
    $indexDropped = false;
    foreach ($data as $field_name => $field_value) {
      $newIndexes = array_merge($newIndexes, [$field_name => $field_value]);
      if (!isset($textFields[$field_name]) && !empty($textIndex)) {
        $this->getCollection()->dropIndex($textIndex);
        $indexDropped = true;
      }
    }

    // If there is already a field indexed and we don't have it now for indexing
    // then drop the complete index
    if (!$indexDropped) {
      foreach (array_keys($textFields) as $field_name) {
        if (!isset($newIndexes[$field_name])) {
          $this->getCollection()->dropIndex($textIndex);
          $indexDropped = true;
        }
      }
    }

    if (empty($newIndexes) && !$indexDropped && !empty($textIndex)) {
      $this->getCollection()->dropIndex($textIndex);
      $indexDropped = true;
    }

    if ((!empty($newIndexes) && $indexDropped) || empty($textIndex)) {
      $this->getCollection()->createIndex($newIndexes);
    }
  }

  protected function compile() {
    $this->condition->compile(null);
    return $this;
  }

  protected function result() {
    // If any field is being added to the entity definition,
    // pretend we don't have any data and return [] otherwise, it would
    // stop the action. We are dealing with noSQL which is schemaless,
    // in this case, Drupal should not care about data stored previously
    if ($_POST && (isset($_POST['op']) && $_POST['op'] === 'Save field settings')) {
      return [];
    }

    /** @var \MongoDB\Collection $collection */
    $collection = $this->getCollection();

    if ($this->count) {
      return $collection->countDocuments($this->filters, $this->options);
    }

    $documents = $collection->find($this->filters, $this->options)->toArray();

    /** @var \MongoDB\Model\BSONDocument[] $result */
    $result = array_map(function ($document) {
      if ($document instanceof BSONDocument) {
        $data = $this->serializer->denormalize($document, 'mongo_to_json');
        if (!$this->returnDocuments) {
          return $data['_id']->__toString();
        } else {
          return $data;
        }
      } else {
        if (!$this->returnDocuments) {
          return $document['_id']->__toString();
        } else {
          return $document;
        }
      }
    }, $documents);

    return $result;
  }

  public function fetchAllAssoc($records, $key) {
    $return = [];
    foreach ($records as $record) {
      $record_key = is_object($record) ? $record->$key : $record[$key];
      $return[$record_key->value] = $record;
    }

    return $return;
  }


  /**
   * {@inheritdoc}
   */
  public function range($start = NULL, $length = NULL) {
    $this->count = null;

    if ($start) {
      $this->range['start'] = (int)$start;
    }

    if ($length) {
      $this->range['length'] = (int)$length;
    }

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function count($flag = TRUE) {
    $this->range = null;

    $this->count = $flag;
    return $this;
  }


  /**
   * @return \MongoDB\Collection
   */
  public function getCollection(): \MongoDB\Collection {
    if (is_null($this->collection)) {
      // TODO: move database into settings
      $this->database = $this->factory->get('entities');
      $this->collection = $this->database->selectCollection($this->getEntityTypeId());
    }
    return $this->collection;
  }

  protected function setOptions() {
    $options = [
      'skip' => isset($this->range['start']) ? $this->range['start'] : null,
      'limit' => isset($this->range['length']) ? $this->range['length'] : null,
      'typeMap' => ['root' => 'array', 'document' => 'array']
    ];

    $this->options = array_merge($this->options, $options);

    if (!is_null($this->projection)) {
      $this->options['projection'] = $this->projection;
    }

    return $this;
  }
}

==> mcet-1.1.x-dev.tar.gz: mcet/src/Query/Query.php <==
<?php

/**
 * @file
 * Contains \Drupal\external_entities\Entity\Query\External\Query.
 */

namespace Drupal\mcet\Query;

use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\Query\QueryBase;
use Drupal\Core\Entity\Query\QueryInterface;
use Drupal\mcet\Entity\MongoContentEntityBaseInterface;
use Drupal\mongodb\DatabaseFactory;
use MongoDB\Model\BSONDocument;
use MongoDB\Model\IndexInfo;
use MongoDB\Operation\Explain;
use MongoDB\Operation\Explainable;
use Symfony\Component\Serializer\Serializer;
use MongoDB\Driver\Exception\CommandException;

/**
 * The SQL storage entity query class.
 */
class Query extends QueryBase implements QueryInterface {

  /** @var \Drupal\mongodb\DatabaseFactory $mongodbFactory */
  public $factory;

  /** @var MongoDb\Database Database */
  public $database;

  /** @var \MongoDB\Collection */
  private $collection = null;

  /** @var \Symfony\Component\Serializer\Serializer $serializer */
  public $serializer;

  public $filters = [];
  protected $options = [];
  protected $projection = [];

  public $fieldDefinitions = [];

  /** @var MongoContentEntityBaseInterface null */
  public $entityInstance = null;
  public $entity_id = null;
  public $bundle = null;

  public $returnDocuments = FALSE;

  /**
   * Constructs a query object.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type definition.
   * @param string $conjunction
   *   - AND: all of the conditions on the query need to match.
   * @param array $namespaces
   *   List of potential namespaces of the classes belonging to this query.
   * @param DatabaseFactory $factory
   *   List of potential namespaces of the classes belonging to this query.*
   */
  public function __construct(EntityTypeInterface $entity_type, $conjunction, array $namespaces, DatabaseFactory $factory, Serializer $serializer) {
    parent::__construct($entity_type, $conjunction, $namespaces);
    $this->factory = $factory;
    $this->serializer = $serializer;
  }


  /**
   * {@inheritdoc}
   */
  public function aggregate($stage, $logic, $langcode = NULL, &$alias = NULL) {
    $this->aggregate[($stage . "_" . sizeof($this->aggregate))] = [
      'stage' => $stage,
      'logic' => $logic,
      'langcode' => $langcode,
    ];

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function execute() {
    return $this
      ->prepare()
      ->compile()
      ->addSort()
      ->result();
  }

  /**
   * @param Explainable $operation
   * @return array|\MongoDB\Collection|object
   */
  public function explain($operation) {
    return $this->getCollection()->explain($operation, $this->options);
  }

  /**
   * @param bool $flag
   * @return $this
   */
  public function setExplain(bool $flag, $verbosity = Explain::VERBOSITY_QUERY): QueryAggregate {
    $this->options['explain'] = $flag;
    $this->options['verbosity'] = $verbosity;
    return $this;
  }

  public function setProjection(array $projection): Query {
    $this->projection = $projection;
    return $this;
  }

  /**
   * @param $filter
   * @param array|null $options
   * @return int
   */
  public function deleteMany($filter, array $options = []) {
    $result = $this->getCollection()->deleteMany($filter, $options);
    return $result->getDeletedCount();
  }

  /**
   * @param array|null $options
   * @return array|object
   */
  public function drop(array $options = []) {
    return $this->getCollection()->drop($options);
  }

  protected function addSort() {
    if ($this->count) {
      $this->sort = [];
    }

    $sort = [];
    if ($this->sort) {
      foreach ($this->sort as $key => $data) {
        $sort[$data['field']] = $data['direction'] === 'DESC' ? -1 : 1;
      }
    }

    $this->options = array_merge($this->options, [
      'sort' => $sort
    ]);

    return $this;
  }

  public function fetchDocuments() {
    $this->returnDocuments = TRUE;
//    $this->projection = $projection;
    return $this;
  }

  protected function prepare() {
    $this->setOptions();

    $this->entity_id = $this->getEntityTypeId();
    $this->bundle = $this->getEntityTypeId();

    $this->fieldDefinitions = \Drupal::entityManager()->getFieldDefinitions($this->entity_id, $this->bundle);

    $entityClass = $this->entityType->getClass();
    $this->entityInstance = new $entityClass([], $this->entity_id, $this->bundle);

    if (method_exists($this->entityInstance, 'getIndexes')) {
      $this->createIndexes();
    }

    return $this;
  }

  protected function createIndexes() {
    $cid = $this->getEntityTypeId() . '_mongo_indexed';
    $cache = \Drupal::cache()->get($cid);
    if ($cache)
      return;

    $requiredIndexes = $this->entityInstance->getIndexes();
    if (is_null($requiredIndexes))
      return;

    $cursor = $this->getCollection()->listIndexes();

    /** @var IndexInfo[] $indexes */
    $indexes = [];
    foreach ($cursor as $index) {
      $indexes[] = $index;
    }

    try {
      foreach ($requiredIndexes as $type => $data) {
        // Produces a single text index as per mongo specifications
        if ($type === 'string_indexes') {
          $this->_indexTextFields($data, $indexes);
        }

        // Produces multiple index per field
        if ($type === 'numeric_indexes') {
          $this->_indexNumericFields($data, $indexes);
        }
      }
    } catch (CommandException $e) {
      _wefiq_log_message("Error on indexes: " . $e->getMessage());
    }

    // TODO:: enable this
    \Drupal::cache()->set($cid, 1);
  }

  /**
   * @param $fields
   * @param \MongoDB\Model\IndexInfo[] $indexes
   */
  private function _indexNumericFields($fields, $indexes) {
    $newIndexes = [];
    foreach ($fields as $field_name => $field_values) {
      $existingIndexes = array_filter(array_map(function ($index) use ($field_name, $field_values) {
        /** @var \MongoDB\Model\IndexInfo $index */

        $fields = $index->getKey();
        $field_keys = array_keys($fields);
        if (in_array($field_name, $field_keys)) {
          $this->getCollection()->dropIndex($index);
          return $index;
        }
      }, $indexes), function ($value) {
        return !is_null($value);
      });

      $newIndexes[] = array_merge([
        'key' => $this->_buildNumericIndexFieldValue($field_name, $field_values)
      ]);
    }

    // drop all indexes which don't exist anymore
    foreach ($indexes as $index) {
      // TODO:: delete old indexes which don't exit anymore
    }

    if (!empty($newIndexes)) {
      $this->getCollection()->createIndexes($newIndexes);
    }
  }

  private function _buildNumericIndexFieldValue($name, $value) {
    $fieldValue = array_merge([
      $name => isset($value['value']) ? $value['value'] : 1,
    ], isset($value['extra']) ? $value['extra'] : []);

    return $fieldValue;
  }

  private function _indexTextFields($data, $indexes) {
    /** @var \MongoDB\Model\IndexInfo $textIndex */
    $textIndex = array_filter(array_map(function ($index) use ($data, $indexes) {
      /** @var \MongoDB\Model\IndexInfo $index */

      $keys = array_keys($index->getKey());

      // TODO: check it in more detail
      if (in_array($keys[0], ['_fts', '_ftx'])) {
        return $index;
      }
    }, $indexes), function ($value) {
      return !is_null($value);
    });

    if (!empty($textIndex)) {
      $textIndex = array_values($textIndex)[0];
    }

    $textFields = [];
    if (!empty($textIndex)) {
      $textFields = $textIndex->offsetGet('weights');
    }

    $newIndexes = [];
    $indexDropped = false;
    foreach ($data as $field_name => $field_value) {
      $newIndexes = array_merge($newIndexes, [$field_name => $field_value]);
      if (!isset($textFields[$field_name]) && !empty($textIndex)) {
        $this->getCollection()->dropIndex($textIndex);
        $indexDropped = true;
      }
    }

    // If there is already a field indexed and we don't have it now for indexing
    // then drop the complete index
    if (!$indexDropped) {
      foreach (array_keys($textFields) as $field_name) {
        if (!isset($newIndexes[$field_name])) {
          $this->getCollection()->dropIndex($textIndex);
          $indexDropped = true;
        }
      }
    }

    if (empty($newIndexes) && !$indexDropped && !empty($textIndex)) {
      $this->getCollection()->dropIndex($textIndex);
      $indexDropped = true;
    }

    if ((!empty($newIndexes) && $indexDropped) || empty($textIndex)) {
      $this->getCollection()->createIndex($newIndexes);
    }
  }

  protected function compile() {
    $this->condition->compile(null);
    return $this;
  }

  protected function result() {
    // If any field is being added to the entity definition,
    // pretend we don't have any data and return [] otherwise, it would
    // stop the action. We are dealing with noSQL which is schemaless,
    // in this case, Drupal should not care about data stored previously
    if ($_POST && (isset($_POST['op']) && $_POST['op'] === 'Save field settings')) {
      return [];
    }

    /** @var \MongoDB\Collection $collection */
    $collection = $this->getCollection();

    if ($this->count) {
      return $collection->countDocuments($this->filters, $this->options);
    }

    $documents = $collection->find($this->filters, $this->options)->toArray();

    /** @var \MongoDB\Model\BSONDocument[] $result */
    $result = array_map(function ($document) {
      if ($document instanceof BSONDocument) {
        $data = $this->serializer->denormalize($document, 'mongo_to_json');
        if (!$this->returnDocuments) {
          return $data['_id']->__toString();
        } else {
          return $data;
        }
      } else {
        if (!$this->returnDocuments) {
          return $document['_id']->__toString();
        } else {
          return $document;
        }
      }
    }, $documents);

    return $result;
  }

  public function fetchAllAssoc($records, $key) {
    $return = [];
    foreach ($records as $record) {
      $record_key = is_object($record) ? $record->$key : $record[$key];
      $return[$record_key->value] = $record;
    }

    return $return;
  }


  /**
   * {@inheritdoc}
   */
  public function range($start = NULL, $length = NULL) {
    $this->count = null;

    if ($start) {
      $this->range['start'] = (int)$start;
    }

    if ($length) {
      $this->range['length'] = (int)$length;
    }

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function count($flag = TRUE) {
    $this->range = null;

    $this->count = $flag;
    return $this;
  }


  /**
   * @return \MongoDB\Collection
   */
  public function getCollection(): \MongoDB\Collection {
    if (is_null($this->collection)) {
      // TODO: move database into settings
      $this->database = $this->factory->get('entities');
      $this->collection = $this->database->selectCollection($this->getEntityTypeId());
    }
    return $this->collection;
  }

  protected function setOptions() {
    $options = [
      'skip' => isset($this->range['start']) ? $this->range['start'] : null,
      'limit' => isset($this->range['length']) ? $this->range['length'] : null,
      'typeMap' => ['root' => 'array', 'document' => 'array']
    ];

    $this->options = array_merge($this->options, $options);

    if (!is_null($this->projection)) {
      $this->options['projection'] = $this->projection;
    }

    return $this;
  }
}
